// -*- mode:C++ -*-
// File:   Moveto_map.h
// Author: Jussi Pajala (jussip@ee.oulu.fi),Antti Tikanmki (sunday@ee.oulu.fi)
#include <Am/Moveto_map.h>

Moveto_map::Moveto_map()
{
     amap.reset();
     Maptype type = Move; 
     amap.set_type(type);
}
void Moveto_map::reset()
{ 
     amap.reset();
}

void Moveto_map::calculate_map(Vector obj_loc, Vector obj_vel)
{
// THESE ARE SET IN MAP CONSTRUCTOR! :
//   decay     : how much object velocity decreases per time instance
//   radius    : how big is the object or how far to avoid it
//   project   : should map be projected or not
//   avoid     : is map for avoiding or not

     float oc_dist,coll_x,coll_y,my_coll_dist,my_spd,my_dir,wgt = -1.0;
     float obj_cur_spd;
     
     reset(); // reset polargrid ( =fill actionmap with zeros )

     float MAX_TIME = Cons2::max_reach_time();
     float coll_time = 0.0;
     float obj_x = obj_loc.get_x();
     float obj_y = obj_loc.get_y();
     obj_dir     = obj_loc.get_dir();
     obj_dist    = obj_loc.get_len();
     
     if(obj_dist<=radius) obj_dist = radius + 0.01;
     
     obj_vel_dir = obj_vel.get_dir();
     obj_spd     = obj_vel.get_len();

     // save some time by precalculating sine and cosine!
     float cosine = cos(obj_vel_dir);
     float sine   = sin(obj_vel_dir);
     
     oc_dist = 0.0;
     obj_cur_spd = obj_spd; 

     // find the first collision time instance 
     // and object position and map weight at that instance 
     for (int cur_time = 1; cur_time <= MAX_TIME; cur_time++){
         // object distance from start point
         oc_dist += obj_cur_spd;
         // reduce velocity by factoring it with decay
         obj_cur_spd *= decay;
         
         // coordinates at current time
         coll_x       = obj_x + oc_dist * cosine;
         coll_y       = obj_y + oc_dist * sine;
         // calculate how far we are from the object edge
         my_coll_dist = sqrt(coll_x*coll_x + coll_y*coll_y) - radius;
         if( my_coll_dist < 0.0 ) my_coll_dist = 0.0;
         int extra_time = 0;
         if(!avoid && cur_time > 1) extra_time = cur_time * 0.2;
         
	 // required speed to catch up the object at this position and time
         my_spd       = my_coll_dist/(cur_time-extra_time);
         // direction to the position
         my_dir       = atan2(coll_y,coll_x);
                  
         if (( my_spd <= CUR_MAX_SPEED)) {
             coll_time = cur_time;
             wgt = time2weight(coll_time);
             
             // if weight is smaller than 1.0 and 
             // speed smaller than current maximum speed it means that 
             // we have not found the best action, because time steps 
             // are discrete. We can approximate speed to be equal to 
             // current maximum. However, if wgt == 1.0, we are so
             // close to the object that even very slow motion gets us 
             // very close to the object, so speed need not be set to
             // cur_max_spd! (But if map is avoid map, faster speeds
             // are filled with 1.0s in fill_collision_angle)
             if( wgt < 1.0 ) my_spd = CUR_MAX_SPEED;
                          
             amap.add_point(my_dir, my_spd, wgt);
             // stop searching best action, it is already found!
	     break;
         }
         
         // stop looping if object velocity is very small. This prevents
         // needless looping when object is static (only one loop needed)
         // and when object velocity has decreased to very small
	 
         // in fact 'very small speed' could be calculated using
         // object distance and (MAX_TIME-cur_time) to calculate what
         // is the smallest effective speed that makes any difference...
         if(obj_cur_spd < 0.01) break;
     }

     // sign of cross product of obj location and velocity
     int cp_sign = 0;
     

     if(wgt < 0){
//         if(obj_spd > 0.0) 
//             cerr << " MAX_TIME too small, no action found!" << __FILE__
//                  << __LINE__ << endl;
         // calculate weight assuming that object is static, i.e
         // velocity is zero
         my_dir = obj_dir;
         my_spd = CUR_MAX_SPEED;
         coll_time = (obj_dist-radius)/my_spd;
         assert(coll_time > 0.0);
         wgt = time2weight(coll_time);
         if(wgt == 0.0) return;
         amap.add_point(my_dir, my_spd, wgt);
         cp_sign = 0;
     } else {
         float cross_product = obj_loc * obj_vel;
         if(cross_product < 0.0) cp_sign = -1;
         else if( cross_product > 0.0) cp_sign = 1;
         else cp_sign = 0;
     }
     
     assert(obj_dist > radius);

     int last_coll_angle = 
         fill_collision_angle(amap.dir2row(my_dir), 
                              amap.mgn2col(my_spd),
                              coll_time, obj_dist, radius);
     
     if(project){
         if(cp_sign)
             project_dyn(cp_sign, amap.dir2row(my_dir) + cp_sign*last_coll_angle,
                         amap.mgn2col(my_spd), coll_time);
         else {
             // if cp_sign == 0, fill to both directions
             project_dyn(1, amap.dir2row(my_dir) + last_coll_angle, 
                         amap.mgn2col(my_spd), coll_time);
             project_dyn(-1, amap.dir2row(my_dir) - last_coll_angle, 
                         amap.mgn2col(my_spd), coll_time);
         }
     }   
     return;
}
