// File: Be_move.cc
// Author: Antti Tikanmki sunday@ee.oulu.fi
//         Jussi Pajala    jussip@ee.oulu.fi
// Purpose: Calculates needed move Actionmaps

#include "Be_move.h"
#include "icomm.h"
#include "tools.h"
#include <assert.h>
#include <stdlib.h>
const float PLAYER_CER_LIMIT = 0.3;
const float dist_limit = 17.0;


// timegap class is a helper class for actionmap printing
class Time_gap 
{
public:
    Time_gap(int gap_size);
    int ok(int time);
protected:
    int last_ok;
    int gap;
};

Time_gap::Time_gap(int time_gap = 10)
{
    last_ok = 0;
    gap = time_gap;
}

int Time_gap::ok(int time)
{
    if(time-last_ok > gap) {
        last_ok = time;
        return 1;
    }
    return 0;
}


Be_move::Be_move()
{
     my_id = BE_MOVE_ID;        // IDs are defined in icomm.h
     inhibited = NO;            // declared in Module.h
     prev_time = curr_time = 0; //    --- '' ---
#ifdef BE_MOVE_AM_SAVE
     am_save_out.open("data/movemaps.bin");
     if (!am_save_out.good())
          cerr << "AM_SAVE_OUT: opening stream failed!" << endl;
#endif
     if(Const::debug_level_startup()) 
         cerr <<  Const::modulename(my_id) << " CONSTRUCTOR executed" << endl;

                                          
}

Be_move::~Be_move()
{
#ifdef BE_MOVE_AM_SAVE
    am_save_out.close();
#endif
}

void Be_move::Init(EmptySamba *sambapointer)
{     
     // store pointer to control software main class.
     samba = sambapointer;
     ma_me = (Ma_me*)samba->get_pointer( MA_ME_ID );
     ma_ball = (Ma_ball*)samba->get_pointer( MA_BALL_ID );
     ma_defloc = (Ma_defloc*)samba->get_pointer( MA_DEFLOC_ID );
     ma_specloc = (Ma_special_loc*)samba->get_pointer( MA_SPECIAL_LOC_ID );
     ma_opponent = (Ma_opponent*)samba->get_pointer( MA_OPPONENT_ID );
     ma_teammate = (Ma_teammate*)samba->get_pointer( MA_TEAMMATE_ID );
     ma_offside = (Ma_offside*)samba->get_pointer( MA_OFFSIDE_ID );
}

int Be_move::must_calc_teammate(int player_no)
{
     if(teammates.player[player_no].loc.get_len()<dist_limit &&
        teammates.player[player_no].cer > PLAYER_CER_LIMIT &&
        teammates.player[player_no].loc.get_dir() ) 
          return 1;
     else return 0;
}
int Be_move::must_calc_opponent(int player_no)
{
     if(opponents.player[player_no].loc.get_len()<dist_limit &&
        opponents.player[player_no].cer > PLAYER_CER_LIMIT )
          return 1;
     else return 0;
}

void Be_move::invalidate_all()
{
    memset(calculated, 0,NUMBER_OF_MOVENAMES*sizeof(int));
    memset(marker_flag, 0,NUMBER_OF_MTYPES*sizeof(int)); 
    ballmap.reset();
    avoidballmap.reset();
    avoidballfarmap.reset();
    for( int i = 0; i < MAX_SEEN_PLAYERS; i++){
        teammatemap[i].reset();
        opponentmap[i].reset();
    }
    all_teammatesmap.reset();
    all_opponentsmap.reset();
    deflocmap.reset();
    gotospecloc.reset();
    gotoball.reset();
    gotoball_no_opp_avoid.reset();
    gotodefloc.reset();
    gotoavoidoffside.reset();
    
};
Moveto_map *Be_move::read(am_name name)
{
    int i;
    static Vector null_velocity;
    
    check_executed();
    if(calculated[name]) return maps[name];

    static Time_gap ball_timer;
        
    switch(name){
    case AM_MBALL: 
        if(!marker_flag[MA_BALL]){
            ma_ball->read(ball);
            marker_flag[MA_BALL] = 1;
        }
        ballmap.calculate_map(ball.loc, ball.mix_vel);
        maps[name] = &ballmap;
        break;
    case AM_AVOID_BALL:
        if(!marker_flag[MA_BALL]){
            ma_ball->read(ball);
            marker_flag[MA_BALL] = 1;
        }
        avoidballmap.calculate_map(ball.loc, ball.mix_vel);
        //avoidballmap.filter_times_longer_than(15);
        maps[name] = &avoidballmap;
        break;
    case AM_AVOID_BALL_FAR:
        if(!marker_flag[MA_BALL]){
            ma_ball->read(ball);
            marker_flag[MA_BALL] = 1;
        }
        avoidballfarmap.calculate_map(ball.loc, ball.mix_vel);
        //avoidballmap.filter_times_longer_than(15);
        maps[name] = &avoidballfarmap;
        break;
    case AM_MTEAMMATE:
         if(!marker_flag[MA_TEAMMATES]){
              ma_teammate->read(teammates);
              marker_flag[MA_TEAMMATES] = 1;
         }
         for(i=0;i<11;i++){
             if( !must_calc_teammate(i)) continue;
             teammatemap[i].calculate_map(teammates.player[i].loc, 
                                          teammates.player[i].mix_vel);
             all_teammatesmap += teammatemap[i];
        }
        maps[name] = &all_teammatesmap;
         break;
    case AM_MOPPONENT:
         if(!marker_flag[MA_OPPONENTS]){
             ma_opponent->read(opponents);
             marker_flag[MA_OPPONENTS] = 1;
         }
         for(i=0;i<11;i++){
             if(!must_calc_opponent(i)) continue;
             opponentmap[i].calculate_map(opponents.player[i].loc, 
                                          opponents.player[i].mix_vel);
             all_opponentsmap += opponentmap[i];
        }
        maps[name] = &all_opponentsmap;
        break;
    case AM_MDEFLOC:
         if(!marker_flag[MA_DEFLOC]){
              ma_defloc->read(defloc);
              marker_flag[MA_DEFLOC] = YES;
         }
         deflocmap.calculate_map(defloc.loc,null_velocity);
//         deflocmap.filter_speeds_faster_than(0.2);
         maps[name] = &deflocmap;
         break;
    case GOTO_BALL:
         gotoball += *read(AM_MBALL);
         gotoball -= *read(AM_MOPPONENT); 
         gotoball -= *read(AM_MTEAMMATE);
         maps[name] = &gotoball;
         // if introduction mode is on, print gotoball action map
         // once in a second
         if(Cons2::introduction() && ball_timer.ok(curr_time)) {
             gotoball.cpp_send("GOTO_BALL");
         }
         break;
    case GOTO_BALL_NO_OPP_AVOID:
         gotoball_no_opp_avoid += *read(AM_MBALL);
         gotoball_no_opp_avoid -= *read(AM_MTEAMMATE);
         maps[name] = &gotoball_no_opp_avoid;
         break;
    case GOTO_DEFLOC:
         gotodefloc += *read(AM_MDEFLOC);
         gotodefloc -= *read(AM_MOPPONENT); 
         gotodefloc -= *read(AM_MTEAMMATE);
         gotodefloc -= *read(AM_AVOID_BALL);
         gotodefloc.filter_speeds_faster_than(0.6);
         maps[name]=&gotodefloc;
         break;
    case GOTO_SPECLOC:
         if(!marker_flag[MA_SPECLOC]){
             ma_specloc->read(specloc);
             marker_flag[MA_SPECLOC] = YES;
         }
         if(Cons2::me_unum()==GOAL_KEEPER_UNUM){
             // if goalkeeper, calculate specloc to default location
             // we can't just use defloc map here, because default location
             // map calculates map for five meter radius, but here we must
             // go exactly to desired position
             if(!marker_flag[MA_DEFLOC]){
                 ma_defloc->read(defloc);
                 marker_flag[MA_DEFLOC] = YES;
             }
             if(defloc.loc.get_len() > 1.0){
                 //cerr << __FILE__ << __LINE__ << "changed " <<  __DATE__ << endl;
                 gotospecloc.calculate_map(defloc.loc,null_velocity);
             } else gotospecloc.reset();
             
             //gotospecloc.filter_speeds_faster_than(0.1);
         }
         else gotospecloc.calculate_map(specloc.loc,null_velocity);
         gotospecloc -= *read(AM_MOPPONENT); 
         gotospecloc -= *read(AM_MTEAMMATE);
         gotospecloc -= *read(AM_AVOID_BALL);
         gotospecloc.filter_speeds_faster_than(0.4);
         
//         if(Cons2::debug_level() && ball_timer.ok(curr_time)) {
//             gotospecloc.cpp_send("GOTO_SPECLOC");
//         }

         maps[name] = &gotospecloc;
         break;
    case GOTO_AVOID_OFFSIDE:
        if(!marker_flag[MA_OFFSIDE]){
            ma_offside->read(offside);
            marker_flag[MA_OFFSIDE] = YES;
        }
        gotoavoidoffside.calculate_map(offside.loc, null_velocity);
        gotoavoidoffside -= *read(AM_AVOID_BALL_FAR);
        gotoavoidoffside -= *read(AM_MOPPONENT);
        gotoavoidoffside -= *read(AM_MTEAMMATE);
/*        if(Cons2::debug_level() && ball_timer.ok(curr_time)) {
            gotoavoidoffside.cpp_send("GOTO_AVOID_OFFSIDE");
        }
*/      
        maps[name] = &gotoavoidoffside;
        break;
    case GOTO_GOODLOC:;
        cerr << __FILE__ << __LINE__ << endl;
        exit(0);
         break;
    default:
         cerr << "SOMETHING WEIRD HAS HAPPENED! " << __FILE__ << " "
              << __LINE__ << " Exiting... " << endl; exit(1);
    }
    calculated[name] = 1;
    return maps[name];   
}

int Be_move::transform()
{
     return 0;
}
