// file    : Ma_ball.cc   

#include "decls.h"
//#include "Soccer.h"
#include <Ma/Ma_ball.h>

const float FILTER_MIN_DIST = 0.0; // minimum distance where filter can be used
const float filt_alpha = 0.8;

//extern void print_vector_to_err(Vector);

Ma_ball::Ma_ball()
{
     my_id = MA_BALL_ID;
     inhibited = NO;
     exec.cer = sense.cer = 0.0;
     
     ball_state = new Ball_state(Cons2::delay(),
                                 Cons2::ball_decay());
     
#ifdef TEST_SAMBA
     if(Cons2::debug_level_startup()) cerr << "Ma_ball: CONSTRUCTOR executed" << endl;
#endif
}

Ma_ball::~Ma_ball()
{
     if(Cons2::debug_level_exit()) 
          cerr << "Ma_ball: DESTRUCTOR executed" << endl;
}

void Ma_ball::Init(EmptySamba *sambapointer)
{
     samba  = sambapointer;
     sensor = (Sensor*)samba->get_pointer( SENSOR_ID );
     ma_me  = (Ma_me*)samba->get_pointer( MA_ME_ID );
     ac_kick = (Ac_kick*)samba->get_pointer( AC_KICK_ID );
     vs_referee = (Vs_referee*)samba->get_pointer( VS_REFEREE_ID );
     vs_team_audio = (Vs_team_audio*)samba->get_pointer( VS_TEAM_AUDIO_ID );
}

void Ma_ball::compute_curr_pos()
{
    if(ball_sensor.valid){
                
        // calculate global location of the ball.
        // because we know robot's global location we can approximate ball's
        // global location
        g_loc = loc2g_loc(ball_sensor.pos, me_sens.g_loc, me_sens.heading);
        g_loc_calculated = 1;
        
        // calculate global velocity if vector length not zero
        if(ball_sensor.vel.get_len() != 0){
            g_vel = loc2g_loc(ball_sensor.vel, me_sens.g_vel, me_sens.heading);
            g_vel_calculated = 1;
        }
    } else if(exec.cer < 0.6){
                
        vs_team_audio->read(team_audio);
        if(team_audio.ball_valid){
            g_loc = team_audio.ball_g_loc;
            g_loc_calculated = 1;
            g_vel = team_audio.ball_g_vel;
            if(g_vel.get_len() > 0.0) g_vel_calculated = 1;
            exec.cer = sense.cer = 0.7;
            if(Cons2::debug_level()) cerr << "using audio" << endl;
            
      }
        
    }
    
}

void Ma_ball::use_team_audio()
{
//     float x,y;
     //stop using filters because information source is to be changed
     // reset ball velocity, audial information doesn't contain
     // velocity information
     sense.g_vel.reset();
     // set ball location from audial information
     sense.g_loc = team_audio.ball_g_loc;
     // copy certainties
     exec.cer = sense.cer = 0.8;
}

void Ma_ball::read_inputs()
{
    ma_me->read_exec(me_exec);
    ma_me->read_sens(me_sens);
    
    
    vs_referee->read(referee);
    team_audio.ball_valid = 0;
    vs_team_audio->read(team_audio);
    
    ac_kick->no_trigger_read(kicks);
    
    if(!Cons2::delay()){
//        cerr << "Ball marker not tested with no-delay-mode" << endl;
    }

    sensor->read_ball_sensor(ball_sensor);
    
}

void Ma_ball::check_game_state()
{
    if(!referee.valid) return;
    
    if( referee.game_state_changed ){
        // request ball_state to reset it's filters, because ball velocity
        // has dropped to zero!
    }

    // as long as game_state is before_kick_off or after_goal, ball must be
    // in the center of the field
    if(referee.game_state == after_goal 
       || referee.game_state == before_kick_off 
       || referee.game_state == opp_kick_off 
       || referee.game_state == team_kick_off){
        static Vector zero_vector; // resetted in constructor!
        //ball_state->reset();
        g_vel = zero_vector;
        g_loc = zero_vector;
        g_loc_calculated = 1;
        g_vel_calculated = 1;
        exec.cer = sense.cer = 1.0;
        cer_calculated = 1;
    }
    if(referee.game_state != play_on){
        g_vel.reset();
        g_vel_calculated = 1;
    }
}

void Ma_ball::update_ball_state()
{
    ball_state->set_rob_pos(me_sens.g_loc,curr_time);
    
    ball_state->set_curr_time(curr_time);
    if(g_loc_calculated)
        ball_state->set_curr_g_pos(g_loc);
    if(g_vel_calculated){
        ball_state->set_curr_g_vel(g_vel);
    }
    
}

void Ma_ball::calculate_exec_and_sense_state()
{
    ball_state->get_exec_params(exec.g_vel, exec.g_loc);
    exec.loc = g_loc2loc(exec.g_loc, me_exec.g_loc, me_exec.heading);
    exec.vel = g_loc2loc(exec.g_vel, me_exec.g_vel, me_exec.heading);
    exec.mix_vel = vel2mix_vel(exec.g_vel, me_exec.heading);

    ball_state->get_curr_params(sense.g_vel, sense.g_loc);
    sense.loc = g_loc2loc(sense.g_loc, me_sens.g_loc, me_sens.heading);
    sense.vel = g_loc2loc(sense.g_vel, me_sens.g_vel, me_sens.heading);
    sense.mix_vel = vel2mix_vel(sense.g_vel, me_sens.heading);
    
        
//    if(fabs(me_sens.heading-me_exec.heading) < 0.1) cerr << "Me_sens: " << me_sens.heading << " Me_exec: " << me_exec.heading << endl;
}


void Ma_ball::calculate_certainty()
{
    exec.valid = 1;
    sense.valid = 1;

    // check if we have already calculated certainty in 
    // check_game_state function
    if(cer_calculated) return;
    // if ball is seen, certainty is 1.0 for both exec and sense moments
    if(ball_sensor.valid) sense.cer = exec.cer = 1.0;
    // if ball was not seen reduce certainty
    else{
        if(sensor->is_any_valid()
           && (ma_me->read_view_angle()/2.0 > fabs(sense.loc.get_dir())))
            // if estimated to be in front of robot, but it is not
            // seen, drop certainty to zero
            sense.cer = exec.cer = 0.0;
        else {
            // if we are not facing the ball reduce certainty steadily
            exec.cer *= Cons2::cer_factor();
            sense.cer = exec.cer;
        }
    }
}

void Ma_ball::invalidate_all()
{
    exec.valid = 0;
    sense.valid = 0;
    vq_change.valid = 0;
}


void Ma_ball::read(Ball_marker &to)
{
    check_executed();
    to = exec;
}

void Ma_ball::read_vq_change(View_change_cmd &to)
{
    check_executed();
    to = vq_change;
}


int Ma_ball::transform()
{
    cer_calculated = g_loc_calculated = g_vel_calculated = 0;
        
    read_inputs();
    
   if(kicks.valid)
        ball_state->set_target_vel(kicks.target_g_vel);
   compute_curr_pos();
   // check game state to get better information about ball position or
    // velocity
    check_game_state();
  update_ball_state();
  calculate_exec_and_sense_state();
   calculate_certainty();

    

//    if(exec.cer == 1.0){
//        vq_change.vq = VQ_High;
/*        if(exec.loc.get_len() < 2.5 && exec.vel.get_len() < 0.3){
            if(Cons2::get_visual_angle() == VA_Narrow) return 0;
            vq_change.valid = 1;
            vq_change.va = VA_Narrow;
            Cons2::set_visual_angle(VA_Narrow);
            cerr << "Player " << Cons2::me_unum() 
                 << " setting visual to NARROW" << endl;
                 } else*/
/*         if(fabs(exec.loc.get_dir()) > 0.4*M_PI){
            if(Cons2::get_visual_angle() == VA_Wide) return 0;
            vq_change.va = VA_Wide;
            vq_change.valid = 1;
            Cons2::set_visual_angle(VA_Wide);
            cerr << "Player " << Cons2::me_unum() 
                 << " setting visual to WIDE" << endl;
       }
        else if(fabs(exec.loc.get_dir()) < 0.3*M_PI && exec.loc.get_len() > 7.0){
            if(Cons2::get_visual_angle() == VA_Normal) return 0;
            vq_change.va = VA_Normal;
            vq_change.valid = 1;
            Cons2::set_visual_angle(VA_Normal);
            cerr << "Player " << Cons2::me_unum() 
                 << " setting visual to NORMAL" << endl;
       }
        
    } else if(exec.cer < 0.6){
        if(Cons2::get_visual_angle() == VA_Wide) return 0;
        vq_change.vq = VQ_High;
        vq_change.valid = 1;
        vq_change.va = VA_Wide;
        Cons2::set_visual_angle(VA_Wide);
        cerr << "Player " << Cons2::me_unum() 
             << " setting visual to WIDE" << endl;
    }
*/  

//    cerr << "Ball exec speed is: " << exec.g_vel.get_len() << endl;
//    ball_state->print_all_states();
//    if(sensor->is_any_valid())
//        cerr << me_exec.g_loc.get_x() << " " << me_exec.g_loc.get_y()<< " "
//             << me_exec.heading << endl;
    return 0;
    
}
