// File     :  Robot_state.cc
// Author   :  Jussi Pajala
// Date     :  30.7./1997

// 30.7.1997  V.0.5     The first version
// 1.8.1997   V.0.6     One (big) bug fixed
// 20.2.1998  v.1.5     A big change

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>
#include <Ma/Robot_state.h>
#include "tools.h"    // for normalize_angle function

//#define ROBOT_STATE_DEBUG
#ifdef ROBOT_STATE_DEBUG
#include <fstream.h>
ofstream rfout;
#endif
  
const int MAX_TIME_LEAP = 8;

//State ver[20];

// INPUTS
// int adv       : tell's how many time instants execution delay is
// float vel_red : velocity is multiplied by this factor at each simulation
//                 cycle
Robot_state::Robot_state(int system_delay, float vel_red)
{
    
     delay = system_delay;
     NSTATES = delay + 1 + MAX_TIME_LEAP;
     state = new State[NSTATES];
     
     decay = vel_red;
     
     for( int i = 0; i < NSTATES; i++){
         state[i].time = 0;
         state[i].speed = 0.0;
         state[i].speed_dir = 0.0;
         state[i].speed_set = 0;
         state[i].heading = 0.0;
         state[i].heading_set = 0;
         state[i].x = 0.0;
         state[i].y = 0.0;
     }
     curr_time = 0;
#ifdef ROBOT_STATE_DEBUG
    rfout.open("Robot_state_output.txt");
#endif
}

Robot_state::~Robot_state()
{
    delete state;
}

// updates the array so that time instances are in sequence.
// also takes care of movements and coordinate updates
void Robot_state::fill_array(int start = 1)
{
    int i;
    
    for(i = start; i < NSTATES; i++){
        if( state[i].time <= state[i-1].time ){
            // if execution reaches this point, we have found an empty
            // array slot. If this is the case, time should be zero.
            // If time is something else, something wierd has happened...
            if(state[i].time){ 
                //cerr << "i: " << i << " t: " << state[i].time 
                //<< endl;
                //print_all_states();
            }
            //assert(state[i].time == 0);
            state[i].time = state[i-1].time + 1;
            state[i].speed_set = 0;
            state[i].heading_set = 0;
        }
        //if speed is not set for this slot, 
        // approximate it with help of previous speed
        if( !state[i].speed_set ){
            state[i].speed = decay * state[i-1].speed;
            state[i].speed_dir = state[i-1].speed_dir;
        }
        //if heading not set, it same as previous
        if( !state[i].heading_set )
            state[i].heading = state[i-1].heading;
        
        // coordinates can be easily approximated with previous coords and
        // previous speed.
        state[i].x = state[i-1].x 
             + state[i-1].speed * cos(state[i-1].speed_dir);
         state[i].y = state[i-1].y 
             + state[i-1].speed * sin(state[i-1].speed_dir);
    }
    
}


void Robot_state::set_curr_time(int time, int time_running)
{
    int i;
    int old_time = curr_time;
    curr_time = time;
    //cerr << "set_curr_time" << endl;

#ifdef ROBOT_STATE_DEBUG
   rfout << "Old time: " 
         << old_time 
         << endl 
         << "New time:" 
         << time << endl;
   rfout_all_states();
    
#endif        
    fill_array();
#ifdef ROBOT_STATE_DEBUG
   rfout << "After fill_Array()" << endl;
   rfout_all_states();
#endif         
    int gap = curr_time - old_time;
    if(gap < 0){
//        cerr << __FILE__ << " time gap negative!" << endl;
        for(i = 0; i < NSTATES; i++) if(state[i].time != 0) state[i].time += gap;
        gap = 0;
    }
    // special case:
    // robots can turn even when time is not running. 
    if(!gap && !time_running){
        // simulate time by shifting previous time instances backwards
        for( i = 0; i < NSTATES; i++ )
            state[i].time--;
        gap = 1;
    }
    
    if(gap){ // return; // there is nothing to do if time gap is zero!
        if(gap>NSTATES-1){
            // this should be a very special case!
            //cerr << "Initializing Robot_state! " << endl;
            state[0].time = time;
            fill_array();
        }
        else{
            // move current time to first place of the array
            for(i = 0; i+gap < NSTATES; i++)
                state[i] = state[i+gap];
           
            // if gap + delay >= NSTATES, it means that MAX_TIME_LEAP is too 
            // small. This means that server time is at least MAX_TIME_LEAP
            // ahead of the robot time.
            assert( gap + delay < NSTATES );
            
            for(i = NSTATES-gap; i < NSTATES; i++){
                state[i].time = 0;
                state[i].speed = 0.0;
                state[i].speed_dir = 0.0;
                state[i].speed_set = 0;
                state[i].heading = 0.0;
                state[i].heading_set = 0;
                state[i].x = 0.0;
                state[i].y = 0.0;
            }
            assert( state[0].time = curr_time );
            
            // this consumes little bit too much time.
            // more optimal approach would be filling just end of the array here.
            fill_array(NSTATES-gap);
        } //else cerr << "gap was zero in " << __FILE__ << endl;
    }
    
    state[delay + 1].heading_set = 0;
    state[delay + 1].speed_set = 0;
    
}

void Robot_state::set_curr_g_pos(Vector g_pos)
{
    //cerr << "set curr pos" << endl;
 
    int i;
    float gap_x, gap_y;
    gap_x = g_pos.get_x() - state[0].x;
    gap_y = g_pos.get_y() - state[0].y;
    
    //cerr << "gap_x = " << gap_x << " gap_y = " << gap_y << endl;
    
    for(i = 0; i < NSTATES; i++){
        state[i].x += gap_x;
        state[i].y += gap_y;
    }
}

void Robot_state::set_curr_speed(float speed)
{
//     cerr << "set curr speed" << endl;
    float gap = speed - state[0].speed;
    int i;
    for(i = 0; i < NSTATES; i++){
        state[i].speed += gap;
        gap *= decay;
    }
}

void Robot_state::set_curr_heading(float heading)
{
//    cerr << "set curr heading: " << heading << endl;
    
    float gap = heading - state[0].heading;
    int i;
    for( i = 0; i < NSTATES; i++){
        state[i].heading += gap;
    }
//    print_all_states();
}

void Robot_state::set_target_speed(float speed)
{
    //cerr << "target_speed" << endl;
    
    fill_array();
    if(!state[delay+1].heading_set == 0)
        //print_all_states();
        ;
    
    /*cerr << "speed : " << state[delay+1].speed  << endl
         << "speed_set: " << state[delay+1].speed_set << endl
         << "heading: " << state[delay+1].heading << endl
         << "heading_set: " << state[delay+1].heading_set << endl;*/
//    }
     
    assert(state[delay+1].heading_set == 0);
    
    state[delay+1].speed = speed;
    state[delay+1].speed_set = 1;
    state[delay+1].speed_dir = state[delay+1].heading;
    state[delay+1].time = state[0].time + delay + 1;
}

void Robot_state::set_target_d_head(float d_heading)
{
    //cerr << "target_heading" << endl;
#ifdef ROBOT_STATE_DEBUG
   rfout << "Before set tarrget d_head" << endl;
   rfout_all_states();
#endif
    
   fill_array();
   
   if(state[delay+1].speed_set) ;//print_all_states();
    
    
    assert(state[delay+1].speed_set == 0);
    
    state[delay+1].heading += d_heading;
    state[delay+1].heading_set = 1;
    state[delay+1].time = state[0].time + delay + 1;
#ifndef NOPRINTS
    if(fabs(state[delay+1].heading) >= 10.0)
        cerr << __FILE__ << " " << __LINE__ << " heading > 10!" << endl;
#endif
    normalize_angle(state[delay+1].heading);

#ifdef ROBOT_STATE_DEBUG
   rfout << "After set target d_head" << endl;
   rfout_all_states();
#endif    
    
    /*cerr << "speed : " << state[delay+1].speed  << endl
         << "speed_set: " << state[delay+1].speed_set << endl
         << "heading: " << state[delay+1].heading << endl
         << "heading_set: " << state[delay+1].heading_set << endl;*/
}

void Robot_state::get_exec_params(Vector &v, Vector &g_pos, float &heading)
{
//     cerr << "get exec params" << endl;
     
    v.set_dir_len(state[delay].speed_dir, state[delay].speed);
    g_pos.set_xy(state[delay].x, state[delay].y);
    heading = state[delay].heading;
}

void Robot_state::get_curr_params(Vector &v, Vector &g_pos, float &heading)
{
//    cerr << "get curr params" << endl;
//    print_all_states();
    
    v.set_dir_len(state[0].speed_dir, state[0].speed);
    g_pos.set_xy(state[0].x, state[0].y);
    heading = state[0].heading;
}

void Robot_state::print_all_states()
{
     int i;
     for(i = 0; i < NSTATES; i++)
     cout << "i " << i << " t " << state[i].time << " spd " << state[i].speed
          << " spd dir " << state[i].speed_dir 
          << " spd_set " << state[i].speed_set
          << " heading " <<  state[i].heading
          << " headset " << state[i].heading_set
          << " x " << state[i].x
          << " y " << state[i].y << endl;
}

void Robot_state::rfout_all_states()
{
#ifdef ROBOT_STATE_DEBUG
    for(int i = 0; i < NSTATES; i++)
       rfout << "i " << i << " t " << state[i].time << " spd " << state[i].speed
             << " spd dir " << state[i].speed_dir 
             << " spd_set " << state[i].speed_set
             << " heading " <<  state[i].heading
             << " headset " << state[i].heading_set
             << " x " << state[i].x
             << " y " << state[i].y << endl;
#endif
}

    
