// -*- mode:C++ -*-

// Base for soccer robot
// File: interface.h
// created by Jussi Pajala
//  
//	this file was based on Itsuki Noda's sample client.h
//  nowadays only a little fraction of this file is derived from Noda's
//  example client
//
// What is class Interface?
//
// This class is designed to work as an interface (just like the name tells)
// between robot hardware and software. As you could imagine, this 
// interface works only with Soccer Server, where hardware is simulated. 
// The Interface creates connection to the Server and reads keyboard.
// It also takes care of timing problems in control software execution,
// command sending and sensor data receiving. These are the main jobs
// of the interface and each job is also a thread. So there is five
// threads:
// - control software thread
// - receive thread
// - send thread
// - timing thread
// - user interface thread
//
// Threads are _almost_ transparent to the user. The main problem with
// them is use of mutexes (or locks). When control software thread returns
// from Interface.tick(), it is owner of a lock that takes care of sensor
// data reservations. When the mutex is locked no other thread can write
// or read the data. Special CARE MUST BE TAKEN that this mutex is unlocked
// AFTER control software has read newest sensor data to its own data store.
// If it is not unlocked right after the data is read, it prevents sensor 
// data receivance thread from writing new data to interface's data stores. 
// Further, if the mutex is not locked at all, execution of the control
// software is jammed when it next time tries to lock the mutex in
// interface.tick().
//
//
// How to use class Interface?
//
// Using it should be relatively easy. Here is a short "user's quide"
//
// Initialization
//
// Instantiate the interface in the main module of your control software. 
// Then call Interface::init() with arguments which were given to your
// main() function (int argc, char *argv[]). Init creates connection to 
// server computer, whose name was given by user. Also, it creates four 
// extra threads to take care of the jobs discussed above.
// Now interface is ready to rock'n roll :)
//
// Running
// 
// You should call Interface.tick() once in one control software cycle.
// Tick pauses the control thread for so long that execution time for
// each cycle becomes 100ms. If control software exceeds the 100ms
// limit, software announces it to the user, but doesn't quit the program.
// If the limit is exceeded no pausing occurs.
// After tick returns new sensor data (if exists) has to be read to
// control software stores and mutex must be immediately unlocked.
// Interface data stores should not be read after unlocking since
// the data might be corrupted because new data is being
// written simultaneously.
//
// Getting sensor data
//
// Information about objects on the field can be read from interface
// by using different get_... functions. These functions are defined below.
// These get functions are the most chaotic things in the Interface.
// Data is located in different kinds of stores in the interface and
// it is delivered to control software with different styles.
// Some structures that Interface uses are defined in files Sensor.h,
// fieldobject.h and icomm.h. I'm planning to do something about these
// definitions, they may be moved into some other file, which defines
// lots of structures used in varied files of the Samba control software.
// Anyway, data relaying must be done with these functions, no matter
// how confusing they are. Carefully inspect each function and get in
// touch how it likes to put the data in your stores.
// Player data relaying is done by directly filling arrays in control 
// software. That's why get_team, get_uk_ and others need pointer as
// inputs while other get functions need just references. These functions
// assumes that pointers delivered point to arrays large enough to hold
// data of all the players. Be warned that if arrays aren't big enough
// data overflow happens!
//
// Other get functions
// ...

// Although the Interface is strictly attached to soccer, there could
// be other uses for well designed robot control software running 
// with it. For example, if sensor data of a real robot is transformed
// so that it reminds the data of the Soccer Server, it might be possible to 
// construct an interface between real robot and software so that the same 
// control sowtware could be used with Soccer Server and a real robot.
//
// 

#ifndef INTERFACE_H
#define INTERFACE_H

//=====================================================================
// Includes
//=====================================================================

#include <sys/time.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <unistd.h>
#include <pthread.h>
#include "tools.h"
#include "udpif.h"
#include "fieldobject.h"
#include "RoundBuffer.h"
#include "Sensor.h" // for some struct definitions

extern int lock100msowner;
extern int tickdata, tick100ms, send10ms;

//======================================================================
//Parameters
//======================================================================
#define NAMEMAX	1024
#define BUFFERSIZE 4096
#define RECV_INTERVAL_MILLISEC 20
#define DEFAULT_TEAMNAME "CAT_Finland"
#define DEFAULT_SERVER_HOST "localhost"
#define DEFAULT_SERVER_PORT 6000
#define EPS 0.00001
//#define RAD2DEG (float)(180.0 / M_PI) 
//#define DEG2RAD (float)(M_PI / 180.0)


//======================================================================
//Types
//======================================================================

struct server_cmd
{
     char text[BUFFERSIZE];
     int time;
};

//======================================================================
//Class Interface
// this class knows
//    how to send information to server
//    what cmd-line arguments mean
//    what time it is now
//    robot's speed
//    current angle of robot
//======================================================================
class Interface {
private:
     pthread_t sensor_tid, sender_tid,keyboard_read_tid, alarm_tid;
     //signal_handler_tid;
     
     pthread_mutex_t sensor_data_mutex;
     pthread_mutex_t send_data_mutex;
     
     // mutex and condition flags to handle 10ms cycles to send data to server
     pthread_mutex_t mutex_10_ms;
     pthread_cond_t cond_10_ms;
     int flag_10_ms;
   
    // mutex and condition flags to handle 100ms cycles to execute 
    // control software 10 times in a second
    pthread_mutex_t mutex_100_ms;
    pthread_cond_t cond_100_ms;
    int flag_100_ms;
    
     RoundBuffer<server_cmd> *cmd_buffer;
     RoundBuffer<Audioinfo> *audiobuffer;

     int print_sensor_data;  // tell's does user want data to be printed
     float xv,yv;            // estimated x and y velocity
     float curr_speed;       // estimated current speed (computed from xv and yv)
     float angle;            // last measured angle
     float old_angle;        // 
     float delta_angle;      // estimated turn angle since last visual information
                              // estimated current angle = angle+delta_angle
     float x,y,old_x,old_y;  // positions
     int last_updated;        // time when robots state was last updated
     
     Socket sock ;
     char* team ;          // name of our team
     char* serverhost ;    // host's name
     int  serverport ;     // port number
     Side side ;           // playing on the left or right side (on the screen)
     char side_char;       // 'r' or 'l'
     int goalie;           // 1 if player is a goalie else 0
     int req_old_protocol; // 1 if old communication protocol is requested
     int unum ;            // uniform number
     int reconnection;     // set to 1 in Init() if reconnectioning
          
     int bufsize ;         // buffer sizes
     char *recv_buffer ;   // received buffer
     char *move_command;   // sending buffer for moving commands 
                           // (move, dash, kick, catch)
     char *send_buffer ;   // sending buffers for other commands

     SYSTEM_STATE control_system_state; // user's request for system state
          
// variables to measure current time
     int curr_t;                 // estimated time, last_visual_time+time_after_visual
     int last_visual_time;       // time instance the last visual sensor information was got
     int time_after_visual;      // estimated time since last sensor info

     BOOL new_visual_information;// this is set when new visual information is got from sensors
     
// private methods     
     void create_threads();
     void analyze_message();     // analyzes messages sent by the server
     
     void my_get_opt(int argc, char** argv);// reads options
     void usage() ;              // puts usage message into cerr stream

     void estimate_situation();  // estimates position, angle etc
     void compute_situation();   // calculates situation from visual sensor information

     BOOL correct_position(); // calculates robot's position using flag and line information

     enum FIELDOBJECTS which_object(char *, char*, char*, int);//string Objname,string arg1,string arg2,int number); 
     
     int get_lineangle();        // returns robots angle calculated from line information

     int  c_ukfs,  // counters, number of unknown friends
          c_ukopps,  // unknown opponents,
          c_ukps;  // and unknown players

/*     Audioinfo heard_msg[MAX_HEARD_MSG];
     int next_msg_read;
     int next_msg_write;*/
     void put_heard_msg(Audioinfo& msg);
   
     Sense_body sb;
     Line lines[NLINES];     // When robot calls Get_line or Get_flag function its side
     Flag flags[NFLAGS];     // is considered and right line or flag is returned 
     
     Player friends[NFRIENDS];     // coplayers, player's index to this array is his Uniform number
     Player uk_friends[NUKFRIENDS];// max 10 unknown friends ( Unum is unknown )
     Player opponents[NOPPS];       // max 11 known opponents (in order of Unum)
     Player uk_opponents[NUKOPPS];  // max 11 unknown opponents ( Unum is unknown )
     Player unknowns[NUNKNOWNS];   // max 11+10 unknown players (even side unknown)

     FieldObject own_goal;          // no need to explain the following   ???
     FieldObject opp_goal;
     FieldObject ball;

     void get_fieldobject(FieldObject &obj,float &dist, float &dir,
                          float &d_dist, float &d_dir, int &valid );

public:
    void wait_100ms();    
    void launch_100ms_wait();
    
    void wait_10ms();
    void launch_10ms_wait();

    void print_robots(Player *r,int N);
    void print_all_robots_and_exit();
    
    void invalidate_all();      // invalidate all sensor information
    
    Interface(char* teamname = DEFAULT_TEAMNAME,
              char* host = DEFAULT_SERVER_HOST,
              int port = DEFAULT_SERVER_PORT);
    
    ~Interface();
    
    void check_message() ;    // checks if any messages from the server
    void send_next_message();
    
    BOOL init(int argc, char** argv) ;
    void quit();
    void tick();              // waits if necessary and executes necessary methods
    
    // these functions handles mutexes. If lock_sensor_data is called when 
    // other thread has already locked the data, 
    // execution is stopped until data is released
    void lock_sensor_data()  { pthread_mutex_lock  (&sensor_data_mutex); }
    void unlock_sensor_data(){ pthread_mutex_unlock(&sensor_data_mutex); }
    void lock_send_data()    { pthread_mutex_lock  (&send_data_mutex);   } 
    void unlock_send_data()  { pthread_mutex_unlock(&send_data_mutex);   }
    /*void lock_100ms()        { pthread_mutex_lock  (&hundred_ms_mutex);  }
      void unlock_100ms()      { pthread_mutex_unlock(&hundred_ms_mutex);  }
    void lock_ten_ms()       { pthread_mutex_lock  (&ten_ms_mutex);      }
    void unlock_ten_ms()     { pthread_mutex_unlock(&ten_ms_mutex);      }*/
    
// the following methods are services for the robot control software.
// software can request information about measured and estimated
// state of the robot.
// it can also send commands to soccerserver through these services.
    
    inline int get_time() { return last_visual_time + time_after_visual; }
    // the following functions relay information of the fieldobjects to the robot control program
    // all angles are changed to radians (from degrees) and their sign is changed negative
    // players:
     
    // all pointers must point to an array of size MAX_SEEN_PLAYERS
    void get_team   (float *dir, float *dist, float *d_dir, float *d_dist, 
                      float *head, int *Unum, int *valid);
     // all pointers must point to an array of size MAX_SEEN_PLAYERS
     void get_uk_team(float *dir, float *dist, float *d_dir, float *d_dist, 
                      float *head, int *valid);
     // all pointers must point to an array of size MAX_SEEN_PLAYERS
     void get_opp    (float *dir, float *dist, float *d_dir, float *d_dist, 
                      float *head, int *Unum, int *valid);
     // all pointers must point to an array of size MAX_SEEN_PLAYERS
     void get_uk_opp (float *dir, float *dist, float *d_dir, float *d_dist, 
                      float *head, int *valid);
     // all pointers must point to an array of size MAX_SEEN_PLAYERS
     void get_uk     (float *dir, float *dist, float *d_dir, float *d_dist,
                      float *head, int *valid);
     // ball and goals:
     void get_ball    (float &dir, float &dist, float &d_dir, float &d_dist, int &valid);
     void get_opp_goal(float &dir, float &dist, float &d_dir, float &d_dist, int &valid);
     void get_own_goal(float &dir, float &dist, float &d_dir, float &d_dist, int &valid);
     // lines, flags and heard messages:
     void get_line(LINETYPES line,float &dir, float &dist, float &d_dir, float &d_dist, int &valid);
     void get_flag(FLAGTYPES flag,float &dir, float &dist, float &d_dir, float &d_dist, int &valid);
     int get_heard_msg(Audioinfo &msg);
     void get_sense_body(Sense_body &body);
     
     void sensor_data_prints() {
            if(print_sensor_data) print_sensor_data = 0;
            else print_sensor_data = 1;
     }
     void change_control_system_state(SYSTEM_STATE state){ 
          control_system_state = state; }
     SYSTEM_STATE get_user_state(){ return control_system_state; };

     
     inline BOOL is_any_valid() // returns true if there was new sensor info last time sensor were read
          {
               return new_visual_information;
     };
     
     
     inline float get_x(){ // returns estimated current x-coordinate
          switch(side){
          case Side_L:return x;
               break;
               
          case Side_R:return -x;
          default: return x;
               
          }
     };
     
     inline float get_y(){ // returns estimated current y-coordinate
          switch(side){
          case Side_L:return -y;
          case Side_R:return y;
          default:return -y;
          }
     };
     
     inline float get_angle(){ // angle is zero when robot is heading towards opponent goal at the middle of the field
          float tempangle = (int)(angle+delta_angle);
          switch(side){
          case Side_L:tempangle = tempangle; break;
          case Side_R:tempangle = (tempangle - 180); break;
          default: tempangle = tempangle;
          }
          
          while(tempangle > 180) tempangle -=360;
          while(tempangle < -180) tempangle +=360;
          return DEG2RAD*tempangle;
     };
     
     inline float get_xv(){ // x-velocity
          return xv;
     }
     inline float get_yv(){ // y-velocity
          return -yv;
     }
     inline int get_uniform_number(){
          return unum;
     }
     
     
     // these functions send commands to the server
     void command_move(int cmd_x, int cmd_y) ;
     void command_turn(float moment) ;
     void command_dash(float power) ;
     void command_kick(float power, float dir) ;
     void command_catch(float dir);// goalie catch ball
     void command_say(char *msg); // msg must point to string max size 500
     void command_sensebody();    // reguest sensebody data from the server
     void command_change_view(ViewAngle va, ViewQuality vq) ;

     // this function allows user to send anything (s)he likes to the server!!
     // using not recommended!
     void command_send_textbuffer(char* sending_buffer); // max size 512
} ;
     
#endif // ifndef INTERFACE_H
